"""
Provides hard-coded values used in this package.  No imports are allowed
since many other modules import this one
"""
__copyright__ = "Copyright 2017, Datera, Inc."

VERSION = "1.2.17"

VERSION_HISTORY = """
Version History:
    1.1.0 -- Initial Version
    1.1.1 -- Metadata Endpoint, Pep8 cleanup, Logging revamp
             Python 3 compatibility, Versioning and Version Headers,
    1.1.2 -- User, Event_system, Internal, Alerts Endpoints/Entities
    1.2.0 -- v2.2 support, API module refactor
    1.2.1 -- Massive rewrite to use /api endpoints instead of manually
             curated list.  Log compression
    1.2.2 -- Change python version check in Connection to be forgivness based
    1.2.4 -- Named object support, hw metics support
    1.2.5 -- v2.2 Paging support
    1.2.6 -- Bugfix for API v2.1 limit parameter
    1.2.7 -- Changed connection to use requests library, added support for
             logs_upload endpoint.  Added 'strict' parameter to constructor
             which allows disabling endpoint validity checks if set to 'False'
    1.2.8 -- Added support for certs, trace_ids and standard logging
    1.2.9 -- Added stream call for use with Metrics style endpoints
    1.2.10 -- Added back create/delete hooks, implemented hook loading and
              a hook inheritance interface.
    1.2.11 -- Increased log rotation size before compression to 50MB
              and decreased number of retained rotations to 5
    1.2.12 -- Added capability of adding custom headers to requests
    1.2.13 -- Added LDAP support to UDC config
    1.2.15 -- Fixed ldap server bug, dropped get_argparser requirement from
              scaffold, changed required requests version because of security
              vulnerability. Updated the readme.
    1.2.16 -- 503 retry support
    1.2.17 -- Bugfix for issue that prevented older API versions from being
              used
"""

API_VERSIONS = ("v2", "v2.1", "v2.2")

REST_PORT = 7717
REST_PORT_HTTPS = 7718

DEFAULT_HTTP_TIMEOUT = 300

PYTHON_2_7_0_HEXVERSION = 0x020700f0
PYTHON_2_7_9_HEXVERSION = 0x020709f0
PYTHON_3_0_0_HEXVERSION = 0x030000f0

CACHED_SCHEMA = "/tmp/.cached-schema"

TIMEOUT_503 = 300



"""
Readers for the /api schema endpoints.  These don't request that endpoint,
they only parse the returned data
"""
import abc
import os
import re

__copyright__ = "Copyright 2017, Datera, Inc."


class EndpointSchema(object):
    def __init__(self, name, path):
        pass

    def methods():
        pass


class SchemaReader(object):

    def __init__(self, api_schema):
        self.schema = api_schema

    @abc.abstractmethod
    def parse(self):
        pass

    @abc.abstractmethod
    def get_endpoint(endpoint):
        pass

    @classmethod
    def get_reader(cls, version):
        return get_reader(version)


class Reader2(SchemaReader):
    ENTITY_RE = re.compile(r"(:\w+)")
    ENTITY_SUB = '_ENT_'
    _endpoint = '/api'

    def __init__(self, api_schema):
        super(Reader2, self).__init__(api_schema)
        self._endpoints = None
        self._entities = None
        self.parse()

    def _fix_last_en(self, path):
        return "".join(path.rsplit(self.ENTITY_SUB, 1)).rstrip("/")

    def _fix_path(self, path):
        return self.ENTITY_RE.sub(self.ENTITY_SUB, path)

    def _normalize(self, path):
        """
        Replaces all sections of the path that are not identified as endpoints
        with the ENTITY_SUB value.  This makes for a normalized view of the
        path so we can reliably compare /api schema paths with their endpoints

        Eg: /app_instances/XXXXXX-XXX-XX-XXXXX/storage_instances
                --> /app_instances/_ENT_/storage_instances
        """
        new_path = path
        bad_parts = [part for part in os.path.normpath(path).split(os.sep)
                     if part not in self._ep_name_set and part != ""]
        for part in bad_parts:
            new_path = new_path.replace(part, self.ENTITY_SUB, 1)
        return new_path

    def parse(self):
        endpoints = []
        entities = []
        # Clean stuff
        pops = []
        for ep in list(self.schema):
            if "(" in ep:
                pops.append(ep)
                body = self.schema[ep]
                newep = ep.split("(")[0]
                body['name'] = newep
                self.schema[newep] = body
                ep = newep
            if not hasattr(self.schema[ep], 'get'):
                pops.append(ep)
            if ":" in ep:
                pops.append(ep)
                body = self.schema[ep]
                newep = self._fix_path(ep)
                if 'name' in body:
                    body['name'] = self._fix_path(body['name'])
                self.schema[newep] = body
                ep = newep
        for pop in pops:
            self.schema.pop(pop)

        for ep, ep_body in self.schema.items():
            epdict = {}
            epdict['name'] = ep.split("/")[-1]
            epdict['path'] = ep
            epdict['children'] = []

            epdict['returns'] = False
            read = ep_body.get('read', None)
            if read and read.get('returnParamself.schema', None):
                epdict['returns'] = True
            if epdict['name'] == self.ENTITY_SUB:
                entities.append(epdict)
            else:
                endpoints.append(epdict)

        root_ep = {'name': 'root',
                   'path': '/',
                   'returns': False,
                   'children': []}
        for endpoint in endpoints:
            parent = "/" + "/".join(
                endpoint['path'].strip("/").split("/")[:-1])
            for other in endpoints:
                if parent == other['path']:
                    other['children'].append(endpoint['name'])
            if len(endpoint['path'].strip("/").split("/")) == 1:
                root_ep['children'].append(endpoint['name'])

        for entity in entities:
            entity['parent_path'] = (
                "/" + "/".join(entity['path'].strip("/").split("/")[:-1]))
            entity['parent'] = entity['parent_path'].split("/")[-1]
            entity['name'] = entity['parent'] + "_entity"
            for endpoint in endpoints:
                parent = "/" + "/".join(
                    endpoint['path'].strip("/").split("/")[:-1])
                if parent == entity['path']:
                    entity['children'].append(endpoint['name'])
        endpoints.append(root_ep)
        self._endpoints = {endpoint['path']: endpoint
                           for endpoint in endpoints}
        self._ep_name_set = {endpoint['name'] for endpoint in endpoints}
        for endpoint in endpoints:
            parts = endpoint['path'].split(os.sep)
            for part in parts:
                if part:
                    self._ep_name_set.add(part)
        for enum in self.get_enums():
            self._ep_name_set.add(enum)

        self._entities = {
            self._fix_last_en(entity['path']): entity for entity in entities}

    def get_enums(self):
        return self.schema[os.path.join('/system/metrics/', self.ENTITY_SUB)][
            'verbs']['read']['urlParamSchema']['metric_name']['enum']

    def get_endpoint(self, path):
        p = self._normalize(path)
        if self.ENTITY_SUB not in p:
            p = self._fix_path(p)
        return self._endpoints.get(p)

    def get_entity(self, path):
        p = self._normalize(path)
        if self.ENTITY_SUB not in p:
            p = self._fix_path(p)
        return self._entities.get(p)


class Reader21(Reader2):
    def __init__(self, api_schema):
        super(Reader21, self).__init__(api_schema)

    def get_enums(self):
        found = []
        found.extend(self.schema[os.path.join('/metrics/io', self.ENTITY_SUB)][
            'read']['urlParamSchema']['metric']['enum'])
        found.extend(self.schema[os.path.join('/metrics/hw', self.ENTITY_SUB)][
            'read']['urlParamSchema']['metric']['enum'])
        return found


class Reader22(Reader21):
    def __init__(self, api_schema):
        super(Reader22, self).__init__(api_schema)


def get_reader(version):
    _readers = {"v2": Reader2,
                "v2.1": Reader21,
                "v2.2": Reader22}
    return _readers[version]



"""
CleanupHandler class which maintains all the cleanup data for complete API
session.
"""
from __future__ import (print_function, absolute_import, unicode_literals,
                        division)

from dfs_sdk.dlogging import get_log
from dfs_sdk.exceptions import ApiError, ApiNotFoundError
from dfs_sdk.hooks.base import BaseHook

LOG = get_log(__name__)

__copyright__ = "Copyright 2018, Datera, Inc."


class CleanupHandler(BaseHook):
    """
    Class which manages all the objects that needs to be cleaned-up.
    """

    def __init__(self):
        """
        Object initialization
        """
        self._cleanup_paths = []  # List of path strings
        self._cleanup_map = {}  # map from path strings to Entity objects

    def supported_versions(self):
        return "v2.1", "v2.2"

    def create_entity(self, entity):
        """ Adding resource to the cleanup map
        """
        if not entity:
            return
        entity_path = entity.get('path', None)
        if not entity_path:
            return
        LOG.debug('Adding entity to cleanup map: %s', entity_path)
        if entity_path in self._cleanup_paths:
            if not isinstance(self._cleanup_map[entity_path], list):
                self._cleanup_map[entity_path] = [
                    self._cleanup_map[entity_path]]
            self._cleanup_paths.append(entity_path)
            self._cleanup_map[entity_path].append(entity)
            return
        self._cleanup_paths.append(entity_path)
        self._cleanup_map[entity_path] = entity

    def delete_entity(self, entity):
        """
        Removing entity from cleanup map
        """
        if not entity:
            return
        entity_path = entity.get('path', None)
        if not entity_path:
            return
        if entity_path not in self._cleanup_paths:
            return
        if isinstance(self._cleanup_map[entity_path], list) and len(
                self._cleanup_map[entity_path]) > 1:
            return
        LOG.debug('Removing entity from cleanup list: %s', entity_path)

        self._cleanup_paths.remove(entity_path)
        if isinstance(self._cleanup_map[entity_path], list):
            for ent in self._cleanup_map[entity_path]:
                if ent.tenant == entity.tenant:
                    index = self._cleanup_map[entity_path].index(ent)
                    break

            del self._cleanup_map[entity_path][index]
            return
        self._cleanup_map.pop(entity_path)

    def prepare_entity(self, entity):
        pass

    def prepare_endpoint(self, endpoint):
        pass

    def cleanup(self):
        """
        Cleanup all resources available in cleanup map. Cleanup will
        happen in reverse order from order in which they created.
        """
        for entity_path in self._cleanup_paths:
            entity_lst = self._cleanup_map[entity_path]
            if isinstance(entity_lst, list):
                for entity in entity_lst:
                    if 'admin_state' in entity:
                        try:
                            if 'tenant' in entity:
                                entity.set(admin_state='offline', force=True,
                                           tenant=entity['tenant'])
                            else:
                                entity.set(admin_state='offline', force=True)
                        except ApiError:
                            LOG.warning(
                                "Could not set entity offline: " + entity_path)
            else:
                entity = entity_lst
                if 'admin_state' in entity:
                    try:
                        if 'tenant' in entity:
                            entity.set(admin_state='offline', force=True,
                                       tenant=entity['tenant'])
                        else:
                            entity.set(admin_state='offline', force=True)
                    except ApiError:
                        LOG.warning(
                            "Could not set entity offline: " + entity_path)

        while self._cleanup_paths:
            entity_path = self._cleanup_paths.pop()  # remove from right
            entity_lst = self._cleanup_map[entity_path]
            if isinstance(entity_lst, list):
                if len(entity_lst) == 1:
                    self._cleanup_map.pop(entity_path)

                for entity in entity_lst:
                    try:
                        LOG.debug('Deleting entity: %s', entity_path)
                        if 'tenant' in entity:
                            entity.delete(tenant=entity['tenant'])
                        else:
                            entity.delete()
                    except ApiNotFoundError:
                        LOG.debug(
                            "Cannot delete entity. Entity does not exist: %s",
                            entity_path)
            else:
                entity_lst = self._cleanup_map.pop(entity_path)
                entity = entity_lst
                try:
                    LOG.debug('Deleting entity: %s', entity_path)
                    if 'tenant' in entity:
                        entity.delete(tenant=entity['tenant'])
                    else:
                        entity.delete()
                except ApiNotFoundError:
                    LOG.debug(
                        "Cannot delete entity. Entity does not exist: %s",
                        entity_path)



"""
Hook loading utilities
"""
from __future__ import (print_function, absolute_import, unicode_literals,
                        division)

import abc

from dfs_sdk.dlogging import get_log

LOG = get_log(__name__)


def load_hooks(api, hooks):
    """
    Primary way to insert hooks into the SDK/API object

    Hooks should inherit from dfs_sdk.hooks.base:BaseHook.
    See dfs_sdk.hooks.cleanup:CleanupHandler for an example
    """
    for hook in hooks:
        context = api.context
        if context.version in hook.supported_versions():
            context.on_entity_create_hooks.append(hook.create_entity)
            context.on_entity_delete_hooks.append(hook.delete_entity)

            context.prepare_entity_hooks.append(hook.prepare_entity)
            context.prepare_endpoint_hooks.append(hook.prepare_endpoint)
        else:
            LOG.warn(
                "Hook does not support API version: {}.  Supported versions: "
                "{}".format(context.version, hook.supported_version()))


class BaseHook(object):

    @abc.abstractmethod
    def supported_versions(self):
        """ Called to check if hook supports current API version.
        Returns: list/tuple of supported API version strings
        Eg: ("v2.1", "v2.2")
        """
        pass

    @abc.abstractmethod
    def create_entity(self, entity):
        """ Called after an entity has been created on the system
        Returns: Modified entity (if any modification)
        """
        pass

    @abc.abstractmethod
    def delete_entity(self, entity):
        """ Called after an entity has been deleted from the system
        Returns: Modified entity (if any modification)
        """
        pass

    @abc.abstractmethod
    def prepare_entity(self, entity):
        """ Called to setup an entity object returned from a REST query
        Returns: Modified entity (if any modification)
        """
        pass

    @abc.abstractmethod
    def prepare_endpoint(self, endpoint):
        """ Called to setup an endpoint object
        Returns: Modified endpoint (if any modification)
        """
        pass



"""
Provides the ApiConnection class
"""
import functools
import io
import json
import os
import random
import threading
import time
import urllib3
import uuid

import requests

from .exceptions import ApiError
from .exceptions import ApiAuthError, ApiConnectionError, ApiTimeoutError
from .exceptions import ApiInternalError, ApiNotFoundError
from .exceptions import ApiInvalidRequestError, ApiConflictError
from .exceptions import Api503BackoffError, Api503RandomError
from .exceptions import Api503RetryError
from .exceptions import ApiValidationFailedError
from .constants import REST_PORT, REST_PORT_HTTPS
from .constants import VERSION, CACHED_SCHEMA, TIMEOUT_503
from .dlogging import get_log
from .schema.reader import get_reader

__copyright__ = "Copyright 2017, Datera, Inc."

LOG = get_log(__name__)

# TODO(_alastor_): Add certificate verification
urllib3.disable_warnings()
retry_type = "backoff"


def _with_authentication(method):
    """
    Decorator to wrap Api method calls so that we login again if our key
    expires.
    """
    @functools.wraps(method)
    def wrapper_method(self, *args, **kwargs):
        """ Call the original method with a re-login if needed """
        # if we haven't logged in yet, log in and then try the method:
        if not self._logged_in:
            LOG.debug("Log in to API...")
            self.login()
            return method(self, *args, **kwargs)
        # already logged in, but retry if needed in case the key expires:
        try:
            args_copy = tuple(args)
            kwargs_copy = dict(kwargs)
            return method(self, *args, **kwargs)
        except ApiAuthError as e:
            if e.message == ('The key provided with the request does not '
                             'correspond to a valid session.'):
                LOG.debug("API auth error, so try to log in again...")
            else:
                LOG.warn("API auth error, so try to log in again...")
            self.login()
            return method(self, *args_copy, **kwargs_copy)
    return wrapper_method


def _with_503_retry(method):
    """
    Decorator to wrap Api method calls so we retry on 503 errors
    """
    @functools.wraps(method)
    def _wrapper_503(self, *args, **kwargs):
        """ Call the original method with backoff """
        tstart = time.time()
        backoff = 0
        while time.time() - tstart < TIMEOUT_503:
            try:
                return method(self, *args, **kwargs)
            except Api503RandomError:
                slp = round(random.random() * 10, 2)
                LOG.warn("Hit 503 Retry.  "
                         "Adding random sleep and trying again in "
                         "{}s".format(slp))
                time.sleep(slp)
            except Api503BackoffError:
                slp = 1 + (backoff * backoff)
                LOG.warn("Hit 503 Retry.  "
                         "Backing off and trying again in {}s".format(slp))
                time.sleep(slp)

        raise
    return _wrapper_503


def _make_unicode_string(inval):
    """
    Converts a string or bytes into a UTF-8 unicode string
    """
    try:
        return unicode(inval, 'utf-8')  # Python2
    except NameError:
        return str(inval, 'utf-8')  # Python3


class ApiConnection(object):
    """
    This class wraps the HTTP connection, translates to/from JSON, and
    handles authentication.
    Its methods raise ApiError (or its subclasses) when things go wrong
    """
    HEADER_DATA = {'Datera-Driver': 'Python-SDK-{}'.format(VERSION)}

    def __init__(self, context):
        """
        Initialize a connection from a context object, which defines
        the hostname, username, password, etc.
        """
        self._context = context
        self._hostname = context.hostname
        self._username = context.username
        self._password = context.password
        self._tenant = context.tenant

        self._version = context.version
        self._cert = context.cert
        self._cert_key = context.cert_key

        self._secure = context.secure
        self._timeout = context.timeout
        self._ldap_server = context.ldap_server
        self._extra_headers = context.extra_headers

        self._lock = threading.Lock()
        self._key = None
        self.reader = None
        self._logged_in = False

    @classmethod
    def from_context(cls, context):
        cls.RETRY_TYPE = context.retry_503_type
        return cls(context)

    def _get_request_attrs(self, urlpath):
        if self._secure:
            protocol = 'https'
            port = REST_PORT_HTTPS
            cert_data = (self._cert, self._cert_key)
        else:
            protocol = 'http'
            port = REST_PORT
            cert_data = None
        api_version = self._version
        host = self._hostname
        connection_string = '{}://{}:{}/v{}/{}'.format(
                protocol, host, port, api_version.strip('v'),
                urlpath.strip('/'))
        return protocol, port, cert_data, api_version, host, connection_string

    def _http_connect_request(self, method, urlpath, headers=None, params=None,
                              body=None, files=None, sensitive=False):
        protocol, port, cert_data, api_version, host, connection_string = \
                self._get_request_attrs(urlpath)
        headers.update(**self._extra_headers)
        request_id = uuid.uuid4()
        if sensitive:
            dbody = "********"
            dheaders = "********"
        else:
            dbody = body
            dheaders = headers
        LOG.debug("\nDatera Trace ID: %(tid)s\n"
                  "Datera Request ID: %(rid)s\n"
                  "Datera Request URL: %(url)s\n"
                  "Datera Request Method: %(method)s\n"
                  "Datera Request Payload: %(payload)s\n"
                  "Datera Request Headers: %(header)s\n",
                  {'tid': getattr(
                      self._context.thread_local, 'trace_id', None),
                   'rid': request_id,
                   'url': connection_string,
                   'method': method,
                   'payload': dbody,
                   'header': dheaders})
        t1 = time.time()
        try:
            resp = getattr(requests, method.lower())(
                    connection_string, headers=headers, params=params,
                    data=body, verify=False, files=files, cert=cert_data)
            if sensitive or '/api' in resp.url:
                payload = "*********"
            else:
                payload = resp.content
            # Python 2/3 compatibility
            try:
                payload = payload.decode('utf-8').replace('\n', '')
            except AttributeError:
                payload = str(payload).replace('\n', '')
            t2 = time.time()
            timedelta = round(t2 - t1, 3)
            LOG.debug("\nDatera Trace ID: %(tid)s\n"
                      "Datera Response ID: %(rid)s\n"
                      "Datera Response TimeDelta: %(delta)ss\n"
                      "Datera Response URL: %(url)s\n"
                      "Datera Response Payload: %(payload)s\n"
                      "Datera Response Object: %(obj)s\n",
                      {'tid': getattr(
                          self._context.thread_local, 'trace_id', None),
                       'rid': request_id,
                       'delta': timedelta,
                       'url': resp.url,
                       'payload': payload,
                       'obj': None})
        except requests.ConnectionError as e:
            raise ApiConnectionError(e, '')
        except requests.Timeout as e:
            raise ApiTimeoutError(e, '')
        if files:
            resp_data = {}
        else:
            resp_data = resp.json()
        resp_status = resp.status_code
        resp_reason = resp.reason
        resp_headers = resp.headers
        self._assert_response_successful(
            method, urlpath, body, resp_data, resp_status, resp_reason)
        return (resp_data, resp_status, resp_reason, resp_headers)

    def _get_schema(self, endpoint):
        """
        Tries to access cached schema, if not available, pulls new schema
        from the remote box.
        """
        data = None
        if os.path.exists(CACHED_SCHEMA):
            with io.open(CACHED_SCHEMA, 'rb') as f:
                fdata = f.read()
                data = {}
                if fdata:
                    try:
                        # Python 2.7
                        data = json.loads(fdata)
                    except TypeError:
                        # Python 3+
                        data = json.loads(fdata.decode('utf-8'))
                if self._version in data:
                    return data[self._version]
                # Making it sensitive so it doesn't clog the logs
                data[self._version] = self.read_endpoint(endpoint,
                                                         sensitive=True)
        else:
            # Making it sensitive so it doesn't clog the logs
            data = {self._version: self.read_endpoint(endpoint,
                                                      sensitive=True)}
        with io.open(CACHED_SCHEMA, 'wb+') as f:
            jdata = json.dumps(data)
            try:
                # Python 2.7
                f.write(jdata)
            except TypeError:
                # Python 3+
                f.write(jdata.encode('utf-8'))
        return data[self._version]

    @_with_503_retry
    def login(self, **params):
        """ Login to the API, store the key, get schema """
        if params:
            if params.get("ldap_server"):
                send_data = {"name": params.get("name"),
                             "password": params.get("password"),
                             "remote_server": params.get("ldap_server")}
            else:
                send_data = {"name": params.get("name"),
                             "password": params.get("password")}
        else:
            send_data = {"name": self._username, "password": self._password}
            if self._ldap_server:
                send_data.update({'remote_server': self._ldap_server})

        body = json.dumps(send_data)

        headers = {}
        headers["content-type"] = "application/json; charset=utf-8"
        urlpath = "/login"
        method = "PUT"
        resp_dict, resp_status, resp_reason, resp_hdrs = \
            self._http_connect_request(method, urlpath, body=body,
                                       headers=headers, sensitive=True)
        if 'key' not in resp_dict or not resp_dict['key']:
            raise ApiAuthError("No auth key returned", resp_dict)
        key = str(resp_dict['key'])
        with self._lock:
            self._key = key
            self._logged_in = True
            Reader = get_reader(self._version)
            reader = Reader(self._get_schema(Reader._endpoint))
            self.reader = reader

    def logout(self):
        """ Perform logout operation with the key"""
        with self._lock:
            key = self._key
            self._key = None
            self._logged_in = False
        headers = dict()
        headers["content-type"] = "application/json; charset=utf-8"
        headers["auth-token"] = key
        urlpath = "/logout"
        method = "PUT"
        self._http_connect_request(method, urlpath, headers=headers)

    @property
    def auth_key(self):
        return self._key

    @auth_key.setter
    def auth_key(self, new_key):
        self._key = new_key

    def _assert_response_successful(self, method, urlpath, body,
                                    resp_data, resp_status, resp_reason):
        """
        Raises an exception if the response was an error
          resp_data (str)
          resp_status (str)
          resp_reason (str)
        """
        if resp_status >= 200 and resp_status <= 299:
            return
        msg = "[REQUEST]: " + method + " " + urlpath + "\n"
        if body is not None:
            msg += str(body) + "\n"
        if resp_data:
            msg += '[RESPONSE]:\n'
            msg += str(resp_data) + "\n"
        msg += str(resp_status) + " " + str(resp_reason)
        if resp_status == 401:
            raise ApiAuthError(msg, resp_data)
        elif resp_status == 404:
            raise ApiNotFoundError(msg, resp_data)
        elif resp_status == 400 or resp_status == 405 or \
                resp_status == 403:
            raise ApiInvalidRequestError(msg, resp_data)
        elif resp_status == 422:
            raise ApiValidationFailedError(msg, resp_data)
        elif resp_status == 409:
            raise ApiConflictError(msg, resp_data)
        elif resp_status == 500:
            raise ApiInternalError(msg, resp_data)
        elif resp_status == 503:
            if self._context.retry_503_type == "backoff":
                raise Api503BackoffError(msg, resp_data)
            elif self._context.retry_503_type == "random":
                raise Api503RandomError(msg, resp_data)
            else:
                raise Api503RetryError(msg, resp_data)
        else:
            raise ApiError(msg, resp_data)

    def _do_request(self, method, urlpath, files=None, data=None, params=None,
                    **kwargs):
        """
        Handle the aggregation of different pages for the request
        """
        resp_meta, resp_data = \
            self._do_auth_request(method, urlpath, data=data, params=params,
                                  files=files, **kwargs)

        # No metadata means no pagination
        if "metadata" not in resp_meta:
            return resp_meta, resp_data

        if "offset" in resp_meta["metadata"]:
            offset = resp_meta["metadata"]["offset"]
        else:
            offset = 0

        # Should we handle POST or PUT?
        # Limiting the method to GET for now
        if (method == "GET" and "total_count" in resp_meta["metadata"] and
                "request_count" in resp_meta["metadata"]):
            # Keep firing until we get everything
            while True:
                ccount = offset + resp_meta["metadata"]["request_count"]
                tcount = resp_meta["metadata"]["total_count"]
                limit = params.get("limit", tcount) if params else tcount
                # If we've hit the limit or there isn't a limit parameter in
                # the returned metadata, we're done
                if limit <= ccount or "limit" not in resp_meta["metadata"]:
                    break

                offset += resp_meta["metadata"]["limit"]
                next_offset = {"offset": offset}
                resp_meta, resp_data_container = \
                    self._do_auth_request(
                        method, urlpath, data, params=next_offset)
                # Expanding the result
                if isinstance(resp_data, dict):
                    resp_data.update(resp_data_container)
                else:
                    resp_data.extend(resp_data_container)
        return resp_meta, resp_data

    def _build_body_headers(self, data=None, files=None, params=None,
                            **kwargs):
        headers = {}
        # tenant header
        if self._version == 'v2':
            pass  # v2 did not support multi-tenancy
        else:
            if self._tenant:
                headers["tenant"] = self._tenant
            else:
                headers["tenant"] = '/root'
            if isinstance(data, dict):
                if 'tenant' in data:
                    headers["tenant"] = data['tenant']
                    data.pop('tenant')
            elif isinstance(params, dict):
                if 'tenant' in params:
                    headers["tenant"] = params['tenant']
                    params.pop('tenant')
        # Auth-Token header
        if self._key:
            headers["Auth-Token"] = self._key
        if not files:
            # content-type header
            headers["content-type"] = "application/json; charset=utf-8"
        if data is None:
            body = None
        elif isinstance(data, str):
            body = data
        elif files is not None:
            # Requests can't push a string and a file at the same time, so we
            # just pass the data as a normal python dict
            body = data
        else:
            body = json.dumps(data)
        return body, headers

    @_with_503_retry
    @_with_authentication
    def _do_auth_request(self, method, urlpath, files=None, data=None,
                         params=None, **kwargs):
        """
        Translates to/from JSON as needed, calls _http_connect_request()
        Bubbles up ApiError on error
        """
        sensitive = kwargs.get('sensitive')
        body, headers = self._build_body_headers(
            data=data, params=params, files=files)
        parsed_data, resp_status, resp_reason, _resp_headers = \
            self._http_connect_request(method, urlpath, params=params,
                                       body=body, headers=headers, files=files,
                                       sensitive=sensitive)

        ret_metadata = {}
        ret_data = parsed_data
        if self._version == 'v2':
            # v2 had no metadata
            ret_data = parsed_data
            ret_metadata = {}
        else:
            if isinstance(parsed_data, dict) and 'data' in parsed_data:
                ret_data = parsed_data.pop('data')
                ret_metadata = parsed_data
        return ret_metadata, ret_data

    @_with_503_retry
    @_with_authentication
    def _do_stream_request(self, urlpath, data=None, params=None, **kwargs):
        body, headers = self._build_body_headers(data=data, params=params)
        try:
            while True:
                parsed_data, resp_status, resp_reason, resp_headers = \
                    self._http_connect_request(
                        "GET", urlpath, params=params, body=body,
                        headers=headers, **kwargs)
                ret_data = parsed_data.pop('data')
                ret_metadata = parsed_data
                yield ret_metadata, ret_data
        except KeyboardInterrupt:
            LOG.debug("Ctrl-C recieved, ending stream")
    ########################################

    def create_entity(self, path, data, sensitive=False):
        """
        Returns the parsed response data
        Raises ApiError on error
        Parameters:
          path (str) - Endpoint path, e.g. "/app_templates"
          data (dict) - e.g. {"name": "myapptemplate"}
        """
        _metadata, data = self._do_request("POST", path, data=data,
                                           sensitive=sensitive)
        return data

    def read_endpoint(self, path, params=None, sensitive=False):
        """
        Returns the parsed response data
        Raises ApiError on error
        Parameters:
          path (str) - Endpoint path, e.g. "/app_templates"
          params (dict) - Querry Params, e.g. "/app_templates?key=value"
        """
        _metadata, data = self._do_request("GET", path, params=params,
                                           sensitive=sensitive)
        return data

    def read_entity(self, path, params=None, sensitive=False):
        """
        Returns the parsed response data
        Raises ApiError on error
        Parameters:
          path (str) - Entity path, e.g. "/app_templates/myapptemplate"
          params (dict) - Querry Params, e.g. "/app_templates?key=value"
        """
        _metadata, data = self._do_request("GET", path, params=params,
                                           sensitive=sensitive)
        return data

    def update_endpoint(self, path, data, sensitive=False):
        """
        Returns the parsed response data
        Raises ApiError on error
        Parameters:
          path (str) - Endpoint path
          data (dict)
        """
        _metadata, data = self._do_request("PUT", path, data=data,
                                           sensitive=sensitive)
        return data

    def update_entity(self, path, data, sensitive=False):
        """
        Returns the parsed response data
        Raises ApiError on error
        Parameters:
          path (str) - Entity path, e.g. "/app_templates/myapptemplate"
          data (dict)
        """
        _metadata, data = self._do_request("PUT", path, data=data,
                                           sensitive=sensitive)
        return data

    def upload_endpoint(self, path, files, data, sensitive=False):
        """
        Returns the parsed response data
        Raises ApiError on error
        Parameters:
          path (str) - Entity path, e.g. "/app_templates/myapptemplate"
          data (dict)
          files (list)
          sensitive (boolean)
        """
        _metadata, data = self._do_request("PUT", path, data=data,
                                           files=files, sensitive=sensitive)
        return data

    def stream_endpoint(self, path, data, interval, timeout):
        """
        Streams Endpoint Data
        Raises ApiError on error
        Parameters:
          path (str) - Entity path, e.g. "/app_templates/myapptemplate"
          data (dict)
        """
        if timeout == 0:
            timeout = "inf"
        try:
            for _metadata, data in self._do_stream_request(path, data=data):
                yield _metadata, data
                time.sleep(interval)
                if timeout != "inf":
                    timeout -= interval
                    if timeout <= 0:
                        LOG.debug("Timeout reached, ending stream")
                        return
        except KeyboardInterrupt:
            LOG.debug("Ctrl-C recieved, ending stream")

    def delete_entity(self, path, data=None, sensitive=False):
        """
        Returns the parsed response data
        Raises ApiError on HTTP error
        Parameters:
          path (str) - Entity path, e.g. "/app_templates/myapptemplate"
        """
        _metadata, data = self._do_request("DELETE", path, data=data,
                                           sensitive=sensitive)
        return data



"""
Handles logging setup

Set the DSDK_LOG_CFG environment variable to set a custom path to a JSON
logging config file

Set the DSDK_LOG_STDOUT environment variable to a level (eg: debug, info) to
also send logging to STDOUT at the specified logging level

Logs are rotated after reaching 50 MB (default, configurable in
debug_logging.json) and compressed (gz) after rotation
"""
import gzip
import io
import json
import logging
import logging.config
import logging.handlers
import os
import sys
import shutil


from .constants import PYTHON_2_7_0_HEXVERSION

LOGDIR = 'log_cfg'
DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), LOGDIR)
DEBUG_LOG = os.path.join(DIR, 'debug_logging.json')
INFO_LOG = os.path.join(DIR, 'info_logging.json')
ERROR_LOG = os.path.join(DIR, 'error_logging.json')
DISABLE = 'disable'


LOGS = {'debug': DEBUG_LOG,
        'info': INFO_LOG,
        'error': ERROR_LOG}


def get_log(name):
    log = logging.getLogger(name)
    if sys.hexversion >= PYTHON_2_7_0_HEXVERSION:
        if not log.handlers:
            log.addHandler(logging.NullHandler())
    return log


def setup_logging(disable=False):
    path = os.getenv('DSDK_LOG_CFG', DEBUG_LOG)
    # Allows for disabling all logging config by the library
    # This is meant to prevent messing with any top-level logging configs
    # Using the library
    if disable or path.lower() == DISABLE:
        return
    if path in LOGS:
        path = LOGS[path]
    if not os.path.exists(path):
        io.open(path, 'w+').close()
    with io.open(path) as f:
        j = json.load(f)
        logging.config.dictConfig(j)
    stdout = os.getenv('DSDK_LOG_STDOUT', None)
    if stdout:
        add_stdout(stdout)


def add_stdout(level):
    root = logging.getLogger()
    root.setLevel(getattr(logging, level.upper()))
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(getattr(logging, level.upper()))
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    root.addHandler(ch)


class CompressedRotatingFileHandler(logging.handlers.RotatingFileHandler):
    def __init__(self, filename, **kwargs):
        self.delay = kwargs.get('delay', None)
        backupCount = kwargs.get('backupCount', 0)
        self.backup_count = backupCount
        logging.handlers.RotatingFileHandler.__init__(self, filename, **kwargs)

    def doArchive(self, old_log):
        with io.open(old_log, 'rb') as log, \
                gzip.open(old_log + '.gz', 'wb') as comp_log:
            shutil.copyfileobj(log, comp_log)
        os.remove(old_log)

    def doRollover(self):
        if self.stream:
            self.stream.close()
            self.stream = None
        if self.backup_count > 0:
            for index in range(self.backup_count - 1, 0, -1):
                sfn = "{}.{}.gz".format(self.baseFilename, index)
                dfn = "{}.{}.gz".format(self.baseFilename, index + 1)
                if os.path.exists(sfn):
                    if os.path.exists(dfn):
                        os.remove(dfn)
                    os.rename(sfn, dfn)
        dfn = ".".join((self.baseFilename, "1"))
        if os.path.exists(dfn):
            os.remove(dfn)
        if os.path.exists(self.baseFilename):
            os.rename(self.baseFilename, dfn)
            self.doArchive(dfn)
        if not self.delay:
            self.stream = self._open()


# Adding this to the main logging module so we can specify
# class: 'logging.CompressedRotatingFileHandler' in our dictConfigs
logging.CompressedRotatingFileHandler = CompressedRotatingFileHandler



from __future__ import (print_function, unicode_literals, division,
                        absolute_import)

import argparse
import copy
import io
import json
import os
import pprint
import re
import sys

from dfs_sdk import get_api as _get_api

IPRE_STR = r'(\d{1,3}\.){3}\d{1,3}'
IPRE = re.compile(IPRE_STR)

SIP = re.compile(r'san_ip\s+?=\s+?(?P<san_ip>%s)' % IPRE_STR)
SLG = re.compile(r'san_login\s+?=\s+?(?P<san_login>.*)')
SPW = re.compile(r'san_password\s+?=\s+?(?P<san_password>.*)')
TNT = re.compile(r'datera_tenant_id\s+?=\s+?(?P<tenant_id>.*)')
LDP = re.compile(r'datera_ldap_server\s+?=\s+?(?P<ldap>.*)')

LATEST = "2.2"
FALLBACK = ["2", "2.1"]

UNIX_HOME = os.path.join(os.path.expanduser('~'))
UNIX_CONFIG_HOME = os.path.join(UNIX_HOME, 'datera')
UNIX_SITE_CONFIG_HOME = '/etc/datera'
CONFIG_SEARCH_PATH = [os.getcwd(), UNIX_HOME, UNIX_CONFIG_HOME,
                      UNIX_SITE_CONFIG_HOME]
CONFIGS = [".datera-config", "datera-config", ".datera-config.json",
           "datera-config.json"]
CINDER_ETC = "/etc/cinder/cinder.conf"

EXAMPLE_CONFIG = {"mgmt_ip": "1.1.1.1",
                  "username": "admin",
                  "password": "password",
                  "tenant": "/root",
                  "api_version": "2.2",
                  "ldap": ""}
DATERA_RC = "datrc"

ENV_MGMT = "DAT_MGMT"
ENV_USER = "DAT_USER"
ENV_PASS = "DAT_PASS"
ENV_TENANT = "DAT_TENANT"
ENV_API = "DAT_API"
ENV_LDAP = "DAT_LDAP"

EXAMPLE_RC = """\
# DATERA ENVIRONMENT VARIABLES
{}=1.1.1.1
{}=admin
{}=password
{}=/root
{}=2.2
{}=
""".format(ENV_MGMT, ENV_USER, ENV_PASS, ENV_TENANT, ENV_API, ENV_LDAP)

ENV_HELP = {ENV_MGMT: "Datera management IP address or hostname",
            ENV_USER: "Datera account username",
            ENV_PASS: "Datera account password",
            ENV_TENANT: "Datera tenant ID. eg: SE-OpenStack",
            ENV_API: "Datera API version. eg: 2.2",
            ENV_LDAP: "Datera account LDAP server"}

_CONFIG = {}
_ARGS = None
VERBOSE = False


def _print_envs():
    print()
    print("DATERA ENVIRONMENT VARIABLES")
    print("============================")
    longest = 0
    for key in ENV_HELP:
        if len(key) > longest:
            longest = len(key)
    for key, help in sorted(ENV_HELP.items()):
        buff = " " * (longest - len(key))
        print("{}{} -- {}".format(buff, key, help))
    print()


def _gen_config():
    if _ARGS.gen_config == "json":
        if any((os.path.exists(c) for c in CONFIGS)):
            print("Config file already exists in current directory.  Please "
                  "move or remove it before generating a new one")
            sys.exit(1)
        with io.open(CONFIGS[-1], 'w+', encoding='utf-8') as f:
            # Python2 Compatibility
            try:
                f.write(unicode(json.dumps(
                    EXAMPLE_CONFIG, ensure_ascii=False, indent=4)))
            except NameError:
                json.dump(EXAMPLE_CONFIG, f, indent=4)
    elif _ARGS.gen_config == "shell":
        if os.path.exists(DATERA_RC):
            print("RC file already exists in current directory. Please move "
                  "or remove it before generating a new one")
            sys.exit(1)
        with io.open(DATERA_RC, 'w+') as f:
            f.write(EXAMPLE_RC)


def _search_config():
    for p in CONFIG_SEARCH_PATH:
        for conf in CONFIGS:
            fpath = os.path.join(p, conf)
            if os.path.exists(fpath):
                return fpath


def _check_config(config_file):
    missing = []
    for key in EXAMPLE_CONFIG:
        if key not in _CONFIG:
            missing.append(key)
    if missing:
        raise EnvironmentError(
            "All config options must be specified by config file, environment "
            "variable or CLI argument. Missing config keys: {}, config_file: "
            "{}".format(missing, config_file))


def _defaults():
    if not _CONFIG.get("tenant"):
        _CONFIG["tenant"] = "/root"
    if not _CONFIG.get("api_version"):
        _CONFIG["api_version"] = LATEST
    if not _CONFIG.get("ldap"):
        _CONFIG["ldap"] = ""


def _read_config():
    global _CONFIG
    if _ARGS and _ARGS.config:
        config_file = _ARGS.config
    else:
        config_file = _search_config()
    if not config_file:
        _CONFIG = _read_cinder_conf()
    else:
        with io.open(config_file) as f:
            _CONFIG = json.loads(f.read())
    if _CONFIG is None:
        _CONFIG = {}
    _env_override()
    _cli_override()
    _defaults()
    _check_config(config_file)


def _read_cinder_conf():
    if not os.path.exists(CINDER_ETC):
        return
    data = None
    found_index = 0
    found_last_index = -1
    with io.open(CINDER_ETC) as f:
        for index, line in enumerate(f):
            if '[datera]' == line.strip().lower():
                found_index = index
                break
        for index, line in enumerate(f):
            if '[' in line and ']' in line:
                found_last_index = index + found_index
                break
    with io.open(CINDER_ETC) as f:
        data = "".join(f.readlines()[
            found_index:found_last_index])
    san_ip = SIP.search(data).group('san_ip')
    san_login = SLG.search(data).group('san_login')
    san_password = SPW.search(data).group('san_password')
    tenant = TNT.search(data)
    ldap = LDP.search(data)
    if tenant:
        tenant = tenant.group('tenant_id')
    else:
        tenant = "/root"
    if ldap:
        ldap = ldap.group('ldap')
    else:
        ldap = ""
    return {"mgmt_ip": san_ip,
            "username": san_login,
            "password": san_password,
            "tenant": tenant,
            "api_version": LATEST,
            "ldap": ldap}


def _cli_override():
    if _ARGS is None:
        return
    if _ARGS.hostname:
        _CONFIG["mgmt_ip"] = _ARGS.hostname
    if _ARGS.username:
        _CONFIG["username"] = _ARGS.username
    if _ARGS.password:
        _CONFIG["password"] = _ARGS.password
    if _ARGS.tenant:
        _CONFIG["tenant"] = _ARGS.tenant
    if _ARGS.api_version:
        _CONFIG["api_version"] = _ARGS.api_version
    if _ARGS.ldap:
        _CONFIG["ldap"] = _ARGS.ldap


def _env_override():
    if ENV_MGMT in os.environ:
        _CONFIG["mgmt_ip"] = os.environ[ENV_MGMT]
    if ENV_USER in os.environ:
        _CONFIG["username"] = os.environ[ENV_USER]
    if ENV_PASS in os.environ:
        _CONFIG["password"] = os.environ[ENV_PASS]
    if ENV_TENANT in os.environ:
        _CONFIG["tenant"] = os.environ[ENV_TENANT]
    if ENV_API in os.environ:
        _CONFIG["api_version"] = os.environ[ENV_API]
    if ENV_LDAP in os.environ:
        _CONFIG["ldap"] = os.environ[ENV_LDAP]


def vprint(*args, **kwargs):
    global VERBOSE
    if VERBOSE:
        print(*args, **kwargs)


def get_api(**kwargs):
    _read_config()
    tenant = _CONFIG["tenant"]
    if tenant and "root" not in tenant and tenant != "all":
        tenant = "/root/{}".format(tenant)
    if not tenant:
        tenant = "/root"
    if not _CONFIG["api_version"]:
        version = "v{}".format(LATEST)
    else:
        version = "v{}".format(_CONFIG["api_version"].strip("v"))
    return _get_api(_CONFIG["mgmt_ip"],
                    _CONFIG["username"],
                    _CONFIG["password"],
                    version,
                    tenant=tenant,
                    secure=True,
                    immediate_login=True,
                    remote_server=_CONFIG["ldap"],
                    **kwargs)


def print_config():
    config = copy.deepcopy(_CONFIG)
    config["password"] = "******"
    pprint.pprint(config)


def get_config():
    if not _CONFIG:
        _read_config()
    return copy.deepcopy(_CONFIG)


def get_argparser(add_help=True):
    global _ARGS, VERBOSE
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--api-version",
                        help="Datera API version to use (default={})".format(
                            LATEST))
    parser.add_argument("--hostname", help="Hostname or IP Address of Datera "
                                           "backend")
    parser.add_argument("--username", help="Username for Datera account")
    parser.add_argument("--password", help="Password for Datera account")
    parser.add_argument("--tenant",
                        help="Tenant Name/ID to search under,"
                             " use 'all' for all tenants")
    parser.add_argument("--ldap", help="Datera LDAP authentication server")
    parser.add_argument("--config", help="Config file location")
    parser.add_argument("--print-envs", action="store_true",
                        help="Print supported environment variables")
    parser.add_argument("--gen-config", choices=["json", "shell"],
                        help="Generate example config")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="Enable verbose output")
    args, _ = parser.parse_known_args()
    _ARGS = args
    VERBOSE = args.verbose
    if args.print_envs:
        _print_envs()
        sys.exit(0)
    if args.gen_config:
        _gen_config()
        sys.exit(0)
    return argparse.ArgumentParser(add_help=add_help, parents=[parser])



"""
Provides the ApiContext class
"""
__copyright__ = "Copyright 2017, Datera, Inc."


class ApiContext(object):
    """
    This object is created by the top level API object, and is passed in
    to all endpoints and entities.
    """

    def __init__(self):
        self.connection = None

        self.hostname = None
        self.username = None
        self.password = None
        self.tenant = None

        self.timeout = None
        self.secure = None
        self.version = None
        self.cert = None
        self.cert_key = None

        self._reader = None
        self.strict = True
        self.thread_local = None
        self.ldap_server = None
        self.retry_503_type = None
        self.extra_headers = {}

        self.on_entity_create_hooks = []
        self.on_entity_delete_hooks = []
        self.prepare_entity_hooks = []
        self.prepare_endpoint_hooks = []

    @property
    def reader(self):
        if not self._reader:
            # self.connection.login()
            self._reader = self.connection.reader
        return self._reader

    @staticmethod
    def _call_hooks(obj, hooks):
        """ Calls a series of callbacks """
        if not hooks:
            return obj
        for hook in hooks:
            ret = hook(obj)
            if ret is not None:
                obj = ret  # hooks can modify returned object
        return obj

    def on_entity_create(self, entity):
        """ Called after an entity has been created on the system """
        return self._call_hooks(entity, self.on_entity_create_hooks)

    def on_entity_delete(self, entity):
        """ Called after an entity has been deleted from the system """
        return self._call_hooks(entity, self.on_entity_delete_hooks)

    def prepare_entity(self, entity):
        """ Called to setup an entity object returned from a REST query """
        return self._call_hooks(entity, self.prepare_entity_hooks)

    def prepare_endpoint(self, endpoint):
        """ Called to setup an endpoint object """
        return self._call_hooks(endpoint, self.prepare_endpoint_hooks)



"""
Python package for interacting with the REST interface of a Datera
storage cluster.
"""
from __future__ import (print_function, absolute_import, unicode_literals,
                        division)

import os
# Renaming imports to prevent easy top-level importing
# We want folks to use get_api(), not these object directly
from .api import DateraApi as _DateraApi
from .api import DateraApi21 as _DateraApi21
from .api import DateraApi22 as _DateraApi22

from .constants import API_VERSIONS, CACHED_SCHEMA
from .exceptions import ApiError
from .exceptions import ApiAuthError
from .exceptions import ApiInvalidRequestError
from .exceptions import ApiNotFoundError
from .exceptions import ApiConflictError
from .exceptions import ApiConnectionError
from .exceptions import ApiTimeoutError
from .exceptions import ApiInternalError
from .exceptions import ApiUnavailableError
from .dlogging import setup_logging
from .hooks.base import load_hooks


__copyright__ = "Copyright 2017, Datera, Inc."


# TODO(mss): generate this from version list imported from constants
VERSION_MAP = {"v2": _DateraApi,
               "v2.1": _DateraApi21,
               "v2.2": _DateraApi22}


def get_api(hostname, username, password, version, tenant=None, strict=True,
            hooks=None, refresh=False, **kwargs):
    """Main entrypoint into the SDK

    Returns a DateraApi object
    Parameters:
      hostname (str) - The hostname or VIP
      username (str) - e.g. "admin"
      password (str)
      version (str) - must be in constants.API_VERSIONS
    Optional parameters:
      tenant (str) - Tenant, for v2.1+ API only
      strict (bool) - Force all requests to use HTTPS (default True)
      hooks (list) - List of hook instances use (see dfs_sdk/hooks/base.py)
      refresh (bool) - Re-generate .cached-schema file from /api endpoint
    """
    setup_logging(kwargs.get('disable_log', False))
    if version not in API_VERSIONS:
        raise ValueError(
            "API version {} unsupported by SDK at this time. Supported "
            "versions: {}".format(version, API_VERSIONS))
    if version == "v2" and tenant:
        raise ValueError("API version v2 does not support multi-tenancy")
    api = VERSION_MAP[version](hostname,
                               username=username,
                               password=password,
                               tenant=tenant,
                               strict=strict,
                               **kwargs)
    if hooks:
        load_hooks(api, hooks)
    if refresh:
        try:
            os.remove(CACHED_SCHEMA)
        except OSError:
            pass
    return api


__all__ = ['get_api',
           'ApiError',
           'ApiAuthError',
           'ApiInvalidRequestError',
           'ApiNotFoundError',
           'ApiConflictError',
           'ApiConnectionError',
           'ApiTimeoutError',
           'ApiInternalError',
           'ApiUnavailableError']



"""
Base classes for endpoint and entitys
"""

import json
import collections
import sys
import os
from .exceptions import _ApiResponseError
from .exceptions import SdkEntityNotFound
from .exceptions import SdkEndpointNotFound
from .constants import PYTHON_3_0_0_HEXVERSION

__copyright__ = "Copyright 2017, Datera, Inc."


###############################################################################

def _is_stringtype(value):
    try:
        return isinstance(value, basestring)
    except NameError:
        return isinstance(value, str)


def get_init_func(klass):
    def paste_init(self, *args):
        super(klass, self).__init__(*args)
    return paste_init


def snake_to_camel(name):
    parts = name.split('_')
    return "".join(x.title() for x in parts)

###############################################################################


class Entity(collections.Mapping):
    """
    Entity object

    This is a mapping, so its attributes can be accessed just like a dict
    """
    _name = "base_entity"

    def __init__(self, context, data, name, parent_path, strict=False):
        """
        Parameters:
          context (dateraapi.context.ApiContext)
          data (dict)
        """
        self.context = context
        self._data = data

        # Set self._path:
        if 'path' in data:
            self._path = data['path']
            self._type = self.context.reader.get_entity(
                os.path.dirname(self._path))
            if not self._type:
                self._type = self.context.reader.get_entity(self._path)
                if not self._type and strict:
                    raise SdkEntityNotFound(
                        "/api endpoint did not contain entity: {} or entity:"
                        " {}".format(os.path.dirname(self._path), self._path))
                else:
                    self._type = {'name': self._name}
        else:
            self._path = parent_path
            self._type = {'name': self._name}
        if 'tenant' in data:
            self._tenant = data['tenant']
        else:
            self._tenant = None

        if name:
            self._name = name

        # In Python 2, dicts have has_key(); it was removed in Python 3.
        # So, do the same thing here:
        if sys.hexversion < PYTHON_3_0_0_HEXVERSION:
            self.has_key = self._has_key

    #
    # Implement dict-like interface

    def __len__(self):
        return len(self._data)

    def __getitem__(self, key):
        return self._data[key]

    def __iter__(self):
        return iter(self._data)

    def _has_key(self, key):
        """ True if this entity contains the given key, else False """
        return key in self

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def items(self):
        return self._data.items()

    ######

    def __str__(self):
        """ A human-readable representation of this entity """
        try:
            # Python 2
            return json.dumps(self._data, encoding='utf-8', indent=4)
        except TypeError:
            # Python 3
            return json.dumps(self._data, indent=4)

    def __repr__(self, **kwargs):
        version = self.context.connection._version
        t = snake_to_camel(self._type.get('name', ''))
        if t == "BaseEntity":
            t = self._name
        return "".join(("<", version, " ", t,
                        " ", str(self._path), " at 0x", str(id(self)), ">"))

    def __getattr__(self, attr):
        """ An attempt to give a functioning object back
        if no subendpoint of the requested name exists

        Lookup Resolution Order:
            Function --> Endpoint --> Data Key

        If data key is preferred, use dictionary syntax:
            ai.storage_instances --> Endpoint
            ai['storage_instances'] --> Data Key List
        """
        if attr in self.__dict__:
            return self.__dict__[attr]
        else:
            if attr not in self.context.reader._ep_name_set:
                if attr in self._data:
                    return self._data[attr]
                else:
                    raise SdkEndpointNotFound(
                        "No {} Endpoint found for {}".format(
                            self.context.connection._version, attr))

            klass = type(snake_to_camel(attr), (GenericEndpoint,), {})
            klass.__init__ = get_init_func(klass)
            klass._name = attr
            return klass(self.context, self._path)

    ######
    # TODO(mss): Allow passing string, then determine the endpoint version
    def _set_subendpoint(self, klass):
        """ Create a sub-endpoint of the given endpoint type """
        assert(issubclass(klass, Endpoint))
        subendpoint = klass(self.context, self._path)
        subendpoint = self.context.prepare_endpoint(subendpoint)
        subendpoint_name = klass._name
        setattr(self, subendpoint_name, subendpoint)

    def reload(self, **params):
        """ Load a new instance of this entity from the API """
        if self._tenant:
            params = {'tenant': self._tenant}

        data = self.context.connection.read_entity(self._path, params)
        entity = self.__class__(self.context, data, self._name, self._path)
        entity = self.context.prepare_entity(entity)
        if self._tenant:
            entity._tenant = self._tenant
        return entity

    def set(self, **params):
        """ Send an API request to modify this entity """
        data = self.context.connection.update_entity(self._path, params)
        entity = self.__class__(self.context, data, self._name, self._path)
        entity = self.context.prepare_entity(entity)
        return entity

    def delete(self, **params):
        """ Send an API request to delete this entity """
        data = self.context.connection.delete_entity(self._path, data=params)
        entity = self.__class__(self.context, data, self._name, self._path)
        entity = self.context.prepare_entity(entity)
        entity = self.context.on_entity_delete(entity)
        return entity


###############################################################################


class Endpoint(object):
    """ REST API endpoint
        There should be a corresponding Entity Object created
        Eg, /network
    """

    _name = "base_endpoint"  # Subclass must initialize it
    _entity_cls = Entity

    def __init__(self, context, parent_path):
        """
        Parameters:
          context (dateraapi.context.ApiContext)
        """
        self.context = context
        self._path = None
        if (not parent_path and
                (self._name == "" or self._name == "base_endpoint")):
            self._path = ""  # root endpoint
        else:
            self._path = parent_path + '/' + self._name

    def __repr__(self):
        name = snake_to_camel(self._name)
        if not name:
            name = 'Root'
        path = repr(self._path)
        if path in ("''", '""'):
            path = repr('/')

        return "".join(("<", self.context.connection._version, ".",
                        name, "Ep", " ", path, ">"))

    def __getattr__(self, attr):
        """ An attempt to give a functioning object back
        if no subendpoint of the requested name exists
        """
        if attr in self.__dict__:
            return self.__dict__[attr]
        else:
            if not self.context.connection:
                self.context.connection = self._create_connection(self.context)
                self.context.connection.login()
            if (attr not in self.context.reader._ep_name_set
                    and self.context.strict):
                raise SdkEndpointNotFound(
                    "No {} Endpoint found for {}".format(
                        self.context.connection._version, attr))

            klass = type(snake_to_camel(attr), (GenericEndpoint,), {})
            klass.__init__ = get_init_func(klass)
            klass._name = attr
            return klass(self.context, self._path)

    # TODO(mss): Allow passing string, then determine the endpoint version
    def _set_subendpoint(self, klass):
        """ Create a sub-endpoint of the given endpoint type """
        assert(issubclass(klass, Endpoint))
        subendpoint = klass(self.context, self._path)
        subendpoint = self.context.prepare_endpoint(subendpoint)
        subendpoint_name = klass._name
        setattr(self, subendpoint_name, subendpoint)

    def _get_list(self, _path, data):
        """ Returns a list of objects or strings, depending on the endpoint """
        if isinstance(data, list):
            return [self._prepare_data(value) for value in data]
        elif isinstance(data, dict):
            return [self._prepare_data(value)
                    for value in data.values()]
        else:
            raise ValueError("Unexpected response: " + repr(data))

    def _link_unlink(self, op, entity, tenant=None):
        if _is_stringtype(entity):
            path = entity
        else:
            path = entity._path
        params = dict(op=op, path=path)
        if tenant:
            params['tenant'] = tenant
        data = self.context.connection.update_endpoint(self._path, params)
        return self._new_contained_entity(data)

    def _prepare_data(self, value):
        """
        If the data looks like a entity, create a Entity object from it,
        else just return it unchanged
        """
        if isinstance(value, dict):
            return self._new_contained_entity(value)
        else:
            return value

    def _new_contained_entity(self, data):
        """ Creates an Entity object """
        name = data.get('name')
        if not name:
            name = snake_to_camel(self._name + "_entity")
        entity = self._entity_cls(self.context, data, name, self._path)
        entity = self.context.prepare_entity(entity)
        return entity

    def get(self, *args, **params):
        """
        Get a entity by its ID
        If no ID, return the whole collection
        """
        if len(args) == 0:
            # GET the whole collection
            path = self._path  # Eg. /storage_templates
            data = self.context.connection.read_endpoint(path, params)
            if isinstance(data, dict):
                # Try preparing entities for each key
                try:
                    new_data = {}
                    for key in data:
                        new_data[key] = self._new_contained_entity(data[key])
                    return new_data
                # Fall back to just preparing the whole thing
                # (SingletonEndpoints ususally do this)
                except AttributeError:
                    return self._new_contained_entity(data)
            elif isinstance(data, list):
                return self._get_list(path, data)
            else:
                return data
        elif len(args) == 1:
            # GET a entity in the collection
            entity_id = args[0]
            # /storage_template/MyTemplate
            path = self._path + "/" + entity_id
            data = self.context.connection.read_entity(path, params=params)
            # This should return a single object in a dictionary form
            if isinstance(data, list):
                return self._get_list(path, data)
            return self._new_contained_entity(data)
        else:
            raise TypeError("Too many arguments for get()")


class GenericEndpoint(Endpoint):
    """ Catch-all endpoint for when one isn't defined by a type-spec
    but we still need to access it.  Allows most common operations.
    """

    def create(self, **params):
        """
        Create an entity in this collection

        Keyword arguments define the attributes of the created entity.
        """
        data = self.context.connection.create_entity(self._path, params)
        entity = self._new_contained_entity(data)
        entity = self.context.on_entity_create(entity)
        return entity

    def upload(self, **params):
        """Upload multipart/file data to and endpoint.  Use parameter 'files'
           to specify the files to be uploaded"""
        files = params.pop('files')
        data = self.context.connection.upload_endpoint(
            self._path, files, params)
        return data

    def stream(self, **params):
        """Stream data from the endpoint with specified 'interval' and
           'timeout' values in seconds"""
        interval = params.pop('interval', 0)
        timeout = params.pop('timeout', 0)
        return self.context.connection.stream_endpoint(
            self._path, params, interval, timeout)

    def set(self, **params):
        """Sets the endpoint with list passed"""
        data = self.context.connection.update_endpoint(self._path, params)
        return data

    def list(self, **params):
        """ Return all entities in this collection as a list """
        path = self._path
        try:
            data = self.context.connection.read_endpoint(path, params)
            return self._get_list(path, data)
        except _ApiResponseError as ex:
            if ex.resp_data['message'].startswith("No data at "):
                return []
            raise

    def get(self, *args, **params):
        return super(GenericEndpoint, self).get(*args, **params)

    def add(self, entity, tenant=None):
        return self._link_unlink('add', entity, tenant=tenant)

    def remove(self, entity, tenant=None):
        return self._link_unlink('remove', entity, tenant=tenant)

    def entity_from_path(self, path, **params):
        data = self.context.connection.read_endpoint(path, params)
        return self._new_contained_entity(data)



"""
Provides the DateraApi objects
"""
import threading

from .constants import DEFAULT_HTTP_TIMEOUT, VERSION
from .connection import ApiConnection
from .context import ApiContext
from .base import Endpoint as _Endpoint

__copyright__ = "Copyright 2017, Datera, Inc."


DEFAULT_API_VERSION = "v2.1"


# Wrapper function to help deduplicate all the code we were getting with the
# different api versions with little to no difference in this class
def _api_getter(base):

    class _DateraBaseApi(base):
        """
        Use this object to talk to the REST interface of a Datera cluster
        """

        def __init__(self, hostname, username=None, password=None, **kwargs):
            """
            Parameters:
              hostname (str) - IP address or host name
              username (str) - Username to log in with, e.g. "admin"
              password (str) - Password to use when logging in to the cluster
              tenant (str) - Tenant, or None
              timeout (float) - HTTP connection  timeout.  If None, use system
                                default.
              secure (boolean) - Use HTTPS instead of HTTP, defaults to HTTPS
              immediate_login (bool) - If True, login when this object is
                                       instantiated, else wait to login until
                                       a request is sent
              thread_local (dict) - Thread local dictionary with trace id
              ldap_server (string) - LDAP server
              extra_headers (dict) - Headers to pass along with all requests
            """
            assert self._version is not None

            if not hostname or not username or not password:
                raise ValueError(
                    "hostname, username, and password are required")

            # Create the context object, common to all endpoints and entities:
            kwargs['hostname'] = hostname
            kwargs['username'] = username
            kwargs['password'] = password
            self._kwargs = kwargs
            self._context = None

            immediate_login = kwargs.get('immediate_login', True)
            if immediate_login:
                # Support both ways of specifying ldap server
                lds = kwargs.get('remote_server', None)
                if not lds:
                    lds = kwargs.get('ldap_server', None)
                self.context.connection.login(
                    name=kwargs.get('username'),
                    password=kwargs.get('password'),
                    ldap_server=lds)

            # Initialize sub-endpoints:
            super(_DateraBaseApi, self).__init__(self._context, None)

        @property
        def context(self):
            kwargs = self._kwargs
            tenant = kwargs.get('tenant', None)
            timeout = kwargs.get('timeout', DEFAULT_HTTP_TIMEOUT)
            secure = kwargs.get('secure', True)
            strict = kwargs.get('strict', True)
            cert = kwargs.get('cert', None)
            cert_key = kwargs.get('cert_key', None)
            thread_local = kwargs.get('thread_local', threading.local())
            retry_503_type = kwargs.get('retry_503_type', "backoff")

            # Support both ways of specifying ldap server
            lds = kwargs.get('remote_server', None)
            if not lds:
                lds = kwargs.get('ldap_server', None)
            ldap_server = lds

            extra_headers = kwargs.get('extra_headers', None)
            if not ldap_server:
                ldap_server = kwargs.get('ldap_server', None)
            if not self._context:
                self._context = ApiContext()
                self.__create_context(
                        self._context,
                        kwargs['hostname'],
                        username=kwargs['username'],
                        password=kwargs['password'],
                        tenant=tenant,
                        timeout=timeout,
                        secure=secure,
                        version=self._version,
                        strict=strict,
                        cert=cert,
                        cert_key=cert_key,
                        thread_local=thread_local,
                        ldap_server=ldap_server,
                        extra_headers=extra_headers,
                        retry_503_type=retry_503_type)
            return self._context

        @context.setter
        def context(self, value):
            self._context = value

        # We really don't want this overridden.  It messes with
        # initialization too much, thus the name-mangle
        def __create_context(self, context, hostname, username=None,
                             password=None, tenant=None, timeout=None,
                             secure=True, version=DEFAULT_API_VERSION,
                             strict=True, cert=None, cert_key=None,
                             thread_local=threading.local(),
                             ldap_server=None, extra_headers=None,
                             retry_503_type=None):
            """
            Creates the context object
            This will be attached as a private attribute to all entities
            and endpoints returned by this API.

            Note that this is responsible for creating a connection object,
            which is an attribute of the context object.
            """
            context.version = version

            context.hostname = hostname
            context.username = username
            context.password = password
            context.tenant = tenant

            context.timeout = timeout
            context.secure = secure
            context.strict = strict
            context.cert = cert
            context.cert_key = cert_key
            context.extra_headers = extra_headers
            if not extra_headers:
                context.extra_headers = {
                    'Datera-Driver': 'Python-SDK-{}'.format(VERSION)}
            context.thread_local = thread_local
            context.ldap_server = ldap_server
            context.retry_503_type = retry_503_type
            context.connection = self._create_connection(context)

        def _create_connection(self, context):
            """
            Creates the API connection object used to communicate over REST
            """
            return ApiConnection.from_context(context)

    return _DateraBaseApi


class RootEp(_Endpoint):
    """
    Top-level endoint, the starting point for all API requests
    """
    _name = ""

    def __init__(self, *args):
        super(RootEp, self).__init__(*args)


class DateraApi(_api_getter(RootEp)):

    _version = 'v2'

    def __init__(self, *args, **kwargs):
        super(DateraApi, self).__init__(*args, **kwargs)


class DateraApi21(_api_getter(RootEp)):

    _version = 'v2.1'

    def __init__(self, *args, **kwargs):
        super(DateraApi21, self).__init__(*args, **kwargs)


class DateraApi22(_api_getter(RootEp)):

    _version = 'v2.2'

    def __init__(self, *args, **kwargs):
        super(DateraApi22, self).__init__(*args, **kwargs)



"""
Provides exceptions classes
"""
import json

from .dlogging import get_log

LOG = get_log(__name__)

__copyright__ = "Copyright 2017, Datera, Inc."


class SdkError(Exception):
    """ This is the base class for SDK exceptions

    Exceptions specifically dealing with the SDK should inherit from the class

    Exceptions dealing with API responses should use exceptions that inherit
    from ApiResponseError
    """
    pass


class ApiError(Exception):
    """ This is the base class for exceptions raised by the API """
    pass


###############################################################################


class ApiConnectionError(ApiError):
    """ Error trying to communicate with the server """
    pass


class ApiTimeoutError(ApiConnectionError):
    """ Timeout waiting for a response """
    pass


###############################################################################


class _ApiResponseError(ApiError):
    """
    This is the base class for exceptions raised due to an error returned
    by the REST server.

    The JSON response payload is made available as exception attributes
    """
    message = None
    code = None
    http = None
    name = None

    def __init__(self, msg, resp_data=None):
        super(_ApiResponseError, self).__init__(msg)
        self.msg = msg
        self.resp_data = resp_data or '{}'
        js = {}
        try:
            js = json.loads(self.resp_data)
        except ValueError:
            LOG.error("Invalid json payload from API response!")
        except TypeError as e:
            LOG.error("Object recieved from API response was unexpected type "
                      "error: {}, data: {}".format(e, self.resp_data))

        for attr in js.keys():
            # Intentionally overwrite exp.message since it's deprecated
            # str(exp) will use exp.args instead of exp.message
            if attr == 'message' or getattr(self, attr, None) is None:
                setattr(self, attr, js[attr])


class ApiAuthError(_ApiResponseError):
    """ Error due to wrong username/password """
    pass


class ApiInternalError(_ApiResponseError):
    """ HTTP 500 """
    pass


class ApiUnavailableError(_ApiResponseError):
    """
    HTTP 503, indicating that the REST server is having a transient error
    communicating with the backend.
    This is a temporary error, e.g. during a node failover or node install.
    """
    pass


class ApiInvalidRequestError(_ApiResponseError):
    """ Incorrect parameters or URL """
    pass


class ApiValidationFailedError(ApiInvalidRequestError):
    """ Request failed validation, see message attribute for reason """
    pass


class ApiNotFoundError(_ApiResponseError):
    """ HTTP 404 Not Found """
    pass


class ApiConflictError(_ApiResponseError):
    """ Edit conflict """
    pass


class Api503RetryError(_ApiResponseError):
    """ The system is overloaded, retry in a bit """


class Api503BackoffError(Api503RetryError):
    """ The system is overloaded, retry with backoff sleep in a bit """


class Api503RandomError(Api503RetryError):
    """ The system is overloaded, retry with random sleep in a bit """


###############################################################################

class SdkTypeNotFound(SdkError):
    """ Generic SDK type not found exception """
    pass


class SdkEndpointNotFound(SdkTypeNotFound):
    """ SDK Endpoint type was not found """
    pass


class SdkEntityNotFound(SdkTypeNotFound):
    """ SDK Entity type was not found """
    pass



#!/usr/bin/env python

import io
import os
import re

from setuptools import setup, find_packages

CONSTANTS_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'src', 'dfs_sdk', 'constants.py')

README_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'README.md')

VRE = re.compile("""VERSION = ['"](.*)['"]""")


def get_version():
    with io.open(CONSTANTS_FILE) as f:
        return VRE.search(f.read()).group(1)


def get_readme():
    with io.open(README_FILE) as f:
        return f.read()


version = get_version()

setup(
    name='dfs_sdk',
    version=version,
    description='Datera Fabric Python SDK',
    long_description=get_readme(),
    author='Datera Automation Team',
    author_email='support@datera.io',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={'dfs_sdk': ['log_cfg/*.json']},
    # include_package_data=True,
    install_requires=[
        "requests",
        "six",
        "urllib3",
    ],
    url='https://github.com/Datera/python-sdk/',
    download_url='https://github.com/Datera/python-sdk/tarball/v{}'.format(
        version),
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)
